#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @created: 02.02.2024
# @author: Aleksey Komissarov
# @contact: ad3002@gmail.com

from collections import deque

import heapq

class WeightedValueHeap:
    def __init__(self):
        self.heap = []
    
    def insert(self, weight, value):
        heapq.heappush(self.heap, (weight, value))
    
    def pop(self):
        if self.heap:
            return heapq.heappop(self.heap)
        return None
    
    def peek(self):
        if self.heap:
            return self.heap[0]
        return None

def update(
    fs_tree,
    queue,
    nucl,
    fs_x,
    fs_xp,
    current_cid,
    cid,
    cutoff,
):
    if len(fs_x) > cutoff:
        fs_tree[cid] = (
            cid,
            [nucl],
            fs_x,
            fs_xp,
            None,
            [],
        )
        fs_tree[current_cid][5].append(cid)
        queue.append(cid)
        cid += 1
    return cid


def build_fs_tree_from_sequence(array, starting_seq_, names_, positions_, cutoff):
    fs_tree = {}
    queue = []
    cid = 0
    seq = starting_seq_
    names = names_[::]
    positions = positions_[::]
    children = []
    fs = (cid, [seq], names, positions, None, children)
    queue.append(cid)
    fs_tree[cid] = fs
    cid += 1
    cutoff = cutoff

    while queue:
        qcid = queue.pop()

        fs = fs_tree[qcid]

        fs_a, fs_c, fs_g, fs_t = [], [], [], []
        fs_ap, fs_cp, fs_gp, fs_tp = [], [], [], []

        for ii, pos in enumerate(fs[3]):
            current_cid = fs[0]
            seq = fs[1]
            name = fs[2][ii]

            if pos + 1 == len(array):
                continue

            nucl = array[pos + 1]
            if nucl == "A":
                fs_a.append(name)
                fs_ap.append(pos + 1)
            elif nucl == "C":
                fs_c.append(name)
                fs_cp.append(pos + 1)
            elif nucl == "G":
                fs_g.append(name)
                fs_gp.append(pos + 1)
            elif nucl == "T":
                fs_t.append(name)
                fs_tp.append(pos + 1)
            
        cid = update(
            fs_tree,
            queue,
            "A",
            fs_a,
            fs_ap,
            current_cid,
            cid,
            cutoff,
        )
        cid = update(
            fs_tree,
            queue,
            "C",
            fs_c,
            fs_cp,
            current_cid,
            cid,
            cutoff,
        )
        cid = update(
            fs_tree,
            queue,
            "G",
            fs_g,
            fs_gp,
            current_cid,
            cid,
            cutoff,
        )
        cid = update(
            fs_tree,
            queue,
            "T",
            fs_t,
            fs_tp,
            current_cid,
            cid,
            cutoff,
        )

    return fs_tree


def iter_fs_tree_from_sequence(array, starting_seq_, names_, positions_, cutoff):
    heap = []
    names = names_[::]
    positions = positions_[::]
    fs = (1, names, positions)
    heapq.heappush(heap, (1, fs))
    cutoff = cutoff
    while heap:
        L, fs = heapq.heappop(heap)

        fs_a, fs_c, fs_g, fs_t = [], [], [], []
        fs_ap, fs_cp, fs_gp, fs_tp = [], [], [], []

        for ii, pos in enumerate(fs[2]):
            
            name = fs[1][ii]

            if pos + 1 == len(array):
                continue

            nucl = array[pos + 1]
            if nucl == "A":
                fs_a.append(name)
                fs_ap.append(pos + 1)
            elif nucl == "C":
                fs_c.append(name)
                fs_cp.append(pos + 1)
            elif nucl == "G":
                fs_g.append(name)
                fs_gp.append(pos + 1)
            elif nucl == "T":
                fs_t.append(name)
                fs_tp.append(pos + 1)
            
        if len(fs_a) > cutoff:
            fs = (
                L+1,
                fs_a,
                fs_ap,
            )
            heapq.heappush(heap, (L+1, fs))
            yield fs

        if len(fs_c) > cutoff:
            fs = (
                L+1,
                fs_c,
                fs_cp,
            )
            heapq.heappush(heap, (L+1, fs))
            yield fs

        if len(fs_g) > cutoff:
            fs = (
                L+1,
                fs_g,
                fs_gp,
            )
            heapq.heappush(heap, (L+1, fs))
            yield fs

        if len(fs_t) > cutoff:
            fs = (
                L+1,
                fs_t,
                fs_tp,
            )
            heapq.heappush(heap, (L+1, fs))
            yield fs



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @created: 05.06.2011
# @author: Aleksey Komissarov
# @contact: ad3002@gmail.com

import gzip


def sc_iter_fasta_file(file_name, inmem=False, lower=False):
    """Iter over fasta file."""

    header = None
    seq = []
    if file_name.endswith(".gz"):
        opener = gzip.open
        decoder = lambda x: x.decode("utf8")
    else:
        opener = open
        decoder = lambda x: x

    with opener(file_name) as fh:
        if inmem:
            data = [decoder(x) for x in fh.readlines()]
            decoder = lambda x: x
        else:
            data = fh
        for line in data:
            line = decoder(line)
            if line.startswith(">"):
                if seq or header:
                    sequence = "".join(seq)
                    if lower:
                        sequence = sequence.lower()
                    yield header, sequence
                header = line.strip()[1:]
                seq = []
                continue
            seq.append(line.strip())
        if seq or header:
            sequence = "".join(seq)
            if lower:
                sequence = sequence.lower()
            yield header, sequence


def sc_iter_arrays_fasta_file(fasta_file):
    """Iter over fasta file and yield sequences."""
    for header, sequence in sc_iter_fasta_file(fasta_file):
        yield sequence



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @created: 20.02.2024
# @author: Aleksey Komissarov
# @contact: ad3002@gmail.com


REVCOMP_DICTIONARY = dict(list(zip("ATCGNatcgn~[] ", "TAGCNtagcn~][ ")))

def get_revcomp(sequence):
    """Return reverse complementary sequence.

    >>> complementary('AT CG')
    'CGAT'

    """
    return "".join(
        REVCOMP_DICTIONARY.get(nucleotide, "") for nucleotide in reversed(sequence)
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Tests for the main decomposition algorithm.
"""

import pytest
from ArraySplitter.decompose import (
    get_top1_nucleotide,
    compute_cuts,
    decompose_array,
    decompose_array_iter1,
    decompose_array_iter2,
)


class TestNucleotideSelection:
    """Test nucleotide selection functions."""
    
    def test_get_top1_nucleotide_simple(self):
        """Test finding most frequent nucleotide."""
        assert get_top1_nucleotide("AAATC") == 'A'
        assert get_top1_nucleotide("TTTTC") == 'T'
        assert get_top1_nucleotide("CGCGC") == 'C'
        assert get_top1_nucleotide("GGGAT") == 'G'
    
    def test_get_top1_nucleotide_tie(self):
        """Test behavior with tied frequencies."""
        # With ties, it should return one of them consistently
        result = get_top1_nucleotide("AATTCCGG")
        assert result in ['A', 'T', 'C', 'G']
    
    def test_get_top1_nucleotide_empty(self):
        """Test with empty sequence."""
        with pytest.raises(Exception):
            get_top1_nucleotide("")


class TestCutSequences:
    """Test cut sequence identification."""
    
    def test_split_array_simple(self):
        """Test splitting array by cut sequence."""
        array = "ATCGATCGATCG"
        cut = "CG"
        
        parts = array.split(cut)
        assert parts == ['AT', 'AT', 'AT', '']
    
    def test_split_array_no_cut(self):
        """Test when cut sequence not present."""
        array = "ATATATATAT"
        cut = "CG"
        
        parts = array.split(cut)
        assert parts == ['ATATATATAT']
    
    def test_split_array_edge_cases(self):
        """Test edge cases for cutting."""
        # Cut at beginning
        assert "CGATAT".split("CG") == ['', 'ATAT']
        
        # Cut at end
        assert "ATATCG".split("CG") == ['ATAT', '']
        
        # Multiple consecutive cuts
        assert "CGCGAT".split("CG") == ['', '', 'AT']


class TestComputeCuts:
    """Test optimal cut computation."""
    
    def test_compute_cuts_perfect_repeat(self):
        """Test with perfect tandem repeat."""
        array = "CAGCAGCAGCAGCAG"
        hints = [
            (1, "C", 5),
            (2, "CA", 5),
            (3, "CAG", 5),
        ]
        
        best_hint, best_score, best_parts = compute_cuts(array, hints)
        
        assert best_hint == "CAG"
        assert best_score == 1.0  # Perfect score
        assert len(best_parts) == 5
    
    def test_compute_cuts_imperfect_repeat(self):
        """Test with imperfect repeat."""
        array = "CAGCAGCATCAGCAG"  # One mutation
        hints = [
            (3, "CAG", 4),
            (3, "CAT", 1),
        ]
        
        best_hint, best_score, best_parts = compute_cuts(array, hints)
        
        assert best_hint == "CAG"
        assert best_score < 1.0  # Not perfect due to mutation


class TestDecomposition:
    """Test the main decomposition functions."""
    
    def test_decompose_array_simple_repeat(self):
        """Test decomposition of simple repeat."""
        array = "ATCATCATCATC"
        monomers = decompose_array(array, depth=5, cutoff=1)
        
        # Should identify ATC as monomer
        assert len(monomers) == 4
        assert all(m == "ATC" for m in monomers)
    
    def test_decompose_array_with_incomplete_end(self):
        """Test with incomplete repeat at end."""
        array = "ATCATCATCAT"  # Missing last C
        monomers = decompose_array(array, depth=5, cutoff=1)
        
        # Should handle incomplete end
        assert len(monomers) >= 3
        assert monomers[0] == "ATC"
    
    @pytest.mark.parametrize("repeat_unit,count", [
        ("AG", 10),
        ("CAG", 8),
        ("ATTCC", 5),
    ])
    def test_decompose_array_various_repeats(self, repeat_unit, count):
        """Test with various repeat units."""
        array = repeat_unit * count
        monomers = decompose_array(array, depth=10, cutoff=1)
        
        assert len(monomers) == count
        assert all(m == repeat_unit for m in monomers)
    
    def test_decompose_array_iter1(self):
        """Test first iteration of decomposition."""
        array = "CAGCAGCAGCAG"
        best_cut = "CAG"
        best_parts = ["", "", "", "", ""]
        period = 3
        
        monomers = decompose_array_iter1(array, best_cut, best_parts, period)
        
        assert len(monomers) == 4
        assert all(m == "CAG" for m in monomers)
    
    def test_decompose_array_iter2_no_refinement_needed(self):
        """Test second iteration when no refinement needed."""
        monomers = ["CAG", "CAG", "CAG", "CAG"]
        most_common = "CAG"
        
        refined = decompose_array_iter2(monomers, most_common)
        
        # Should not change perfect monomers
        assert refined == monomers
    
    def test_decompose_array_iter2_with_long_monomer(self):
        """Test second iteration with concatenated monomers."""
        monomers = ["CAG", "CAG", "CAGCAG", "CAG"]  # One double monomer
        most_common = "CAG"
        
        refined = decompose_array_iter2(monomers, most_common)
        
        # Should split the double monomer
        assert len(refined) == 5
        assert refined[2] == "CAG"
        assert refined[3] == "CAG"


class TestEdgeCases:
    """Test edge cases and error conditions."""
    
    def test_empty_array(self):
        """Test with empty array."""
        with pytest.raises(Exception):
            decompose_array("", depth=5, cutoff=1)
    
    def test_single_nucleotide(self):
        """Test with single nucleotide."""
        monomers = decompose_array("A", depth=5, cutoff=1)
        assert monomers == ["A"]
    
    def test_no_repeats(self):
        """Test with non-repetitive sequence."""
        array = "ACGTACGTACGT"
        monomers = decompose_array(array, depth=5, cutoff=5)
        
        # High cutoff should return whole sequence
        assert len(monomers) == 1
    
    def test_very_long_repeat_unit(self):
        """Test with very long repeat unit."""
        repeat_unit = "ATCGATCGATCGATCGATCG"  # 20bp
        array = repeat_unit * 3
        
        monomers = decompose_array(array, depth=25, cutoff=1)
        assert len(monomers) == 3
        assert all(m == repeat_unit for m in monomers)


@pytest.fixture
def satellite_array():
    """Generate a satellite array with some variation."""
    base_monomer = "ATTCCATTCCATTCC"
    # Add some variations
    monomers = [base_monomer] * 8
    monomers[3] = "ATTCCATTCTATTCC"  # Single mutation
    monomers[6] = "ATTCCATTCCATTC"   # Deletion
    return "".join(monomers)


def test_realistic_satellite(satellite_array):
    """Test with realistic satellite array."""
    monomers = decompose_array(satellite_array, depth=20, cutoff=3)
    
    # Should find approximately correct number of monomers
    assert 6 <= len(monomers) <= 10
    
    # Most monomers should be similar
    base = "ATTCCATTCCATTCC"
    similar_count = sum(1 for m in monomers if len(m) >= 14 and len(m) <= 16)
    assert similar_count >= len(monomers) * 0.7  # At least 70% similar length


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @created: 05.06.2011
# @author: Aleksey Komissarov
# @contact: ad3002@gmail.com

"""
Classes:
    
- AbstractBlockFileIO(AbstractFileIO)
    
"""

from PyExp import AbstractFileIO


class AbstractBlockFileIO(AbstractFileIO):
    """Working with file with data organized in block, where each block starts with same token.

    Public methods:

    - get_block_sequence(self, head_start, next_head, fh)
    - get_blocks(self, token, fh)
    - gen_block_sequences(self, token, fh)

    Inherited public properties:

    - data  - iterable data, each item is tuple (head, body)
    - N     - a number of items in data

    Overrided public methods:

    - __init__(self, token)
    - read_from_file(self, input_file)
    - read_online(self, input_file) ~> item

    Inherited public methods:

    - read_from_db(self, db_cursor)
    - write_to_file(self, output_file)
    - write_to_db(self, db_cursor)
    - read_as_iter(self, source)
    - iterate(self) ~> item of data
    - do(self, cf, args) -> result
    - process(self, cf, args)
    - clear(self)
    - do_with_iter(self, cf, args) -> [result,]
    - process_with_iter(self, cf, args)

    """

    def __init__(self, token, **args):
        """Overrided. Set token velue."""
        super(AbstractBlockFileIO, self).__init__(**args)
        self.token = token
        self.wise_opener = self.get_opener()

    def read_from_file(self, input_file):
        """Overrided. Read data from given input_file."""
        with self.wise_opener(input_file, "r") as fh:
            for head, body, start, next in self.gen_block_sequences(self.token, fh):
                self.data.append((head, body, start, next))

    def read_online(self, input_file):
        """Overrided. Yield items from data online from input_file."""
        with self.wise_opener(input_file, "r") as fh:
            for head, body, start, next in self.gen_block_sequences(self.token, fh):
                yield (head, body, start, next)

    def get_block_sequence(self, head_start, next_head, fh):
        """Get a data block (head, seq, head_start, head_end).

        Arguments:

        - head_start -- a head starting position in a file
        - next_head  -- a next to head starting position in a file
        - fh         -- an open file handler

        Return format:

        - head       -- a block head
        - seq        -- a block body
        - head_start -- a file pointer to block start
        - head_end   -- a file pointer to next block start or 0
        """
        head_start = int(head_start)
        next_head = int(next_head)
        sequence = ""
        fh.seek(head_start)
        pos = head_start
        if next_head:
            head = fh.readline()
            pos = fh.tell()
            while pos != next_head:
                temp_seq = fh.readline()
                pos = fh.tell()
                if pos != next_head:
                    sequence += temp_seq
            sequence += temp_seq
        else:
            head = fh.readline()
            fasta_list = fh.readlines()
            sequence = "".join(fasta_list)
        return (head, sequence, head_start, next_head)

    def get_blocks(self, token, fh):
        """Get a list of the token positions in given file (first, next).
        For the last string the function returns (last string, 0).

        Arguments:

        - token -- the token indicating a block start
        - fh    -- an open file handler

        Return format: a list of (start, next start) tuples
        """
        fh.seek(0)
        header_start_list = []
        here = 0
        wrong = 0
        start = -1
        while fh:
            here = fh.tell()
            line = fh.readline()
            if len(line) == 0:
                header_start_list.append((start, 0))
                break
            if line.startswith(token):
                if start == -1:
                    start = here
                else:
                    header_start_list.append((start, here))
                    start = here
        return header_start_list

    def gen_block_sequences(self, token, fh):
        """Yield (head, seq, head_start, head_end) tuplefor given fh for open file.

        Arguments:

        - token -- the token indicating a block start
        - fh    -- an open file handler

        Return format:

        - head       -- a block head
        - seq        -- a block body
        - head_start -- a file pointer to block start
        - head_end   -- a file pointer to next block start or 0

        """
        header_start_list = self.get_blocks(token, fh)
        for x, y in header_start_list:
            # yeild sequence by coordinates
            yield self.get_block_sequence(x, y, fh)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @created: 21.02.2024
# @author: Aleksey Komissarov
# @contact: ad3002@gmail.com

import argparse

from .core_functions.tools.rotation import rotate_arrays

from .core_functions.io.fasta_reader import sc_iter_fasta_file

def main(fasta_file, output_file, starting_kmer=None):

    arrays = []
    for ii, (header, seq) in enumerate(sc_iter_fasta_file(fasta_file)):
        arrays.append(seq)

    new_arrays = rotate_arrays(arrays, starting_kmer=starting_kmer)
    
    with open(output_file, "w") as fw:
        for ii, (header, seq) in enumerate(sc_iter_fasta_file(fasta_file)):
            fw.write(f'>{header}\n{" ".join(new_arrays[ii])}\n')
            
def run_it():
    parser = argparse.ArgumentParser(description='Rotate arrays to start from the same position.')
    parser.add_argument('-i', '--fasta', type=str, help='Fasta file', required=True)
    parser.add_argument('-o', '--output', type=str, help='Output file', required=True)
    parser.add_argument('-s', '--start', type=str, help='Starting kmer for rotation [None]', default=None)
    args = parser.parse_args()
    main(args.fasta, args.output, starting_kmer=args.start)

if __name__ == '__main__':
    run_it()


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Simple tests for sequence utility functions.
"""

import pytest
from ArraySplitter.core_functions.tools.sequences import (
    get_revcomp,
    REVCOMP_DICTIONARY,
)


class TestReverseComplement:
    """Test reverse complement function."""
    
    def test_get_revcomp_simple(self):
        """Test basic reverse complement."""
        assert get_revcomp("ATCG") == "CGAT"
        assert get_revcomp("AAAA") == "TTTT"
        assert get_revcomp("TTTT") == "AAAA"
        assert get_revcomp("CCCC") == "GGGG"
        assert get_revcomp("GGGG") == "CCCC"
    
    def test_get_revcomp_mixed(self):
        """Test with mixed sequences."""
        assert get_revcomp("ATCGATCG") == "CGATCGAT"
        assert get_revcomp("GCTAGCTA") == "TAGCTAGC"
    
    def test_get_revcomp_palindrome(self):
        """Test palindromic sequences."""
        # These should be self-complementary when reversed
        assert get_revcomp("ATCGAT") == "ATCGAT"
        assert get_revcomp("GCATGC") == "GCATGC"
    
    def test_get_revcomp_empty(self):
        """Test empty sequence."""
        assert get_revcomp("") == ""
    
    def test_get_revcomp_lowercase(self):
        """Test handling of lowercase."""
        assert get_revcomp("atcg") == "cgat"
        assert get_revcomp("AtCg") == "CgAt"
    
    @pytest.mark.parametrize("seq,expected", [
        ("A", "T"),
        ("T", "A"),
        ("C", "G"),
        ("G", "C"),
        ("ATCGN", "NCGAT"),
    ])
    def test_get_revcomp_parametrized(self, seq, expected):
        """Parametrized tests for various sequences."""
        assert get_revcomp(seq) == expected
    
    def test_revcomp_idempotent(self):
        """Test that double reverse complement returns original."""
        original = "ATCGATCGATCG"
        revcomp1 = get_revcomp(original)
        revcomp2 = get_revcomp(revcomp1)
        assert revcomp2 == original
    
    def test_revcomp_dictionary(self):
        """Test the REVCOMP_DICTIONARY constant."""
        # Check basic mappings
        assert REVCOMP_DICTIONARY['A'] == 'T'
        assert REVCOMP_DICTIONARY['T'] == 'A'
        assert REVCOMP_DICTIONARY['C'] == 'G'
        assert REVCOMP_DICTIONARY['G'] == 'C'
        
        # Check lowercase
        assert REVCOMP_DICTIONARY['a'] == 't'
        assert REVCOMP_DICTIONARY['t'] == 'a'
        
        # Check ambiguous codes
        assert REVCOMP_DICTIONARY['N'] == 'N'
        assert REVCOMP_DICTIONARY['n'] == 'n'


@pytest.fixture
def test_sequences():
    """Provide test sequences."""
    return {
        "simple": "ATCG",
        "repeat": "ATATATAT",
        "complex": "ATCGATCGATCGATCG",
        "ambiguous": "ATCGNNNNATCG",
    }


def test_revcomp_preserves_length(test_sequences):
    """Test that reverse complement preserves sequence length."""
    for name, seq in test_sequences.items():
        revcomp = get_revcomp(seq)
        assert len(revcomp) == len(seq), f"Length mismatch for {name}"


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @created: 14.06.2011
# @author: Aleksey Komissarov
# @contact: ad3002@gmail.com

import re

from PyExp import AbstractModel

from ..tools.parsers import (parse_chromosome_name,
                                                        parse_fasta_head,
                                                        trf_parse_head,
                                                        trf_parse_line)


def get_gc(sequence):
    """Count GC content."""
    length = len(sequence)
    if not length:
        return 0
    count_c = sequence.count("c") + sequence.count("C")
    count_g = sequence.count("g") + sequence.count("G")
    gc = float(count_c + count_g) / float(length)
    return float(gc)


def clear_sequence(sequence):
    """Clear sequence (full alphabet):

    - lower case
    - \s -> ""
    - [^actgn] -> ""
    """
    sequence = sequence.strip().upper()
    sequence = re.sub("\s+", "", sequence)
    return re.sub("[^actgnuwsmkrybdhvACTGNUWSMKRYBDHV\-]", "", sequence)


class TRModel(AbstractModel):
    """Class for tandem repeat wrapping

    Public properties:

    Indexes:

    - id,
    - giid,
    - trf_id,
    - project,
    - trf_gi

    Coordinates:

    - trf_l_ind,
    - trf_r_indel

    TR:

    - trf_period,
    - trf_n_copy,
    - trf_pmatch,
    - trf_pvar,
    - trf_entropy,
    - trf_array_length,
    - trf_joined,
    - trf_chr

    Sequence:

    - trf_consensus,
    - trf_array

    GC%:

    - trf_array_gc,
    - trf_consensus_gc

    Other:

    - trf_head,
    - trf_params

    Annotation:

    - trf_family,
    - trf_subfamily,
    - trf_subsubfamily,
    - trf_hor,
    - trf_n_chrun,
    - trf_n_refgenome,
    - trf_ref_annotation,
    - trf_bands_refgenome,
    - trf_repbase,
    - trf_strand

    Dumpable attributes:

    - "project",
    - "id" (int),
    - "trf_id" (int),
    - "trf_type",
    - "trf_family",
    - "trf_family_prob",
    - "trf_l_ind" (int),
    - "trf_r_ind" (int),
    - "trf_period" (int),
    - "trf_n_copy" (float),
    - "trf_pmatch" (float),
    - "trf_pvar" (float),
    - "trf_consensus",
    - "trf_array",
    - "trf_array_gc" (float),
    - "trf_consensus_gc" (float),
    - "trf_gi",
    - "trf_head",
    - "trf_param",
    - "trf_array_length" (int),
    - "trf_chr",
    - "trf_joined" (int),
    - "trf_superfamily",
    - "trf_superfamily_ref",
    - "trf_superfamily_self",
    - "trf_subfamily",
    - "trf_subsubfamily",
    - "trf_family_network",
    - "trf_family_self",
    - "trf_family_ref",
    - "trf_hor" (int),
    - "trf_n_chrun" (int),
    - "trf_ref_annotation",
    - "trf_bands_refgenome",
    - "trf_repbase",
    - "trf_strand",

    """

    dumpable_attributes = [
        "project",
        "id",
        "trf_id",
        "trf_type",
        "trf_family",
        "trf_family_prob",
        "trf_l_ind",
        "trf_r_ind",
        "trf_period",
        "trf_n_copy",
        "trf_pmatch",
        "trf_pvar",
        "trf_entropy",
        "trf_consensus",
        "trf_array",
        "trf_array_gc",
        "trf_consensus_gc",
        "trf_gi",
        "trf_head",
        "trf_param",
        "trf_array_length",
        "trf_chr",
        "trf_joined",
        "trf_superfamily",
        "trf_superfamily_ref",
        "trf_superfamily_self",
        "trf_subfamily",
        "trf_subsubfamily",
        "trf_family_network",
        "trf_family_self",
        "trf_family_ref",
        "trf_hor",
        "trf_n_chrun",
        "trf_ref_annotation",
        "trf_bands_refgenome",
        "trf_repbase",
        "trf_strand",
    ]

    int_attributes = [
        "trf_l_ind",
        "trf_r_ind",
        "trf_period",
        "trf_array_length",
        "trf_joined",
        "trf_hor",
        "trf_n_chrun",
    ]

    float_attributes = [
        "trf_n_copy",
        "trf_family_prob",
        "trf_entropy",
        "trf_pmatch",
        "trf_pvar",
        "trf_array_gc",
        "trf_consensus_gc",
    ]

    def set_project_data(self, project):
        """Add project data to self.project."""
        self.project = project

    def set_raw_trf(self, head, body, line):
        """Init object with data from parsed trf ouput.

        Parameters:

        - trf_obj: TRFObj instance
        - head: parsed trf head
        - body: parsed trf body
        - line: parsed trf line
        """
        self.trf_param = 0
        self.trf_head = trf_parse_head(head).strip()
        self.trf_gi = parse_fasta_head(self.trf_head)[0]
        self.trf_chr = parse_chromosome_name(self.trf_head)

        (
            self.trf_l_ind,
            self.trf_r_ind,
            self.trf_period,
            self.trf_n_copy,
            self.trf_l_cons,
            self.trf_pmatch,
            self.trf_indels,
            self.trf_score,
            self.trf_n_a,
            self.trf_n_c,
            self.trf_n_g,
            self.trf_n_t,
            self.trf_entropy,
            self.trf_consensus,
            self.trf_array,
        ) = trf_parse_line(line)

        self.trf_pmatch = float(self.trf_pmatch)
        self.trf_pvar = int(100 - float(self.trf_pmatch))

        try:
            self.trf_l_ind = int(self.trf_l_ind)
        except:
            print("Error:", self)

        self.trf_r_ind = int(self.trf_r_ind)
        self.trf_period = int(self.trf_period)
        self.trf_n_copy = float(self.trf_n_copy)

        self.trf_consensus = clear_sequence(self.trf_consensus)
        self.trf_array = clear_sequence(self.trf_array)

        self.trf_array_gc = get_gc(self.trf_array)
        self.trf_consensus_gc = get_gc(self.trf_consensus)
        self.trf_chr = parse_chromosome_name(self.trf_head)
        self.trf_array_length = len(self.trf_array)

    def get_header_string(self):
        """Get header string for tsv file."""
        data = "\t".join(self.dumpable_attributes)
        return f"#{data}\n"

    def get_fasta_repr(self, add_project=False):
        """Get array fasta representation, head - trf_id."""
        if add_project:
            return ">%s_%s\n%s\n" % (self.trf_id, self.project, self.trf_array)
        else:
            return ">%s\n%s\n" % (self.trf_id, self.trf_array)

    @property
    def fasta(self):
        return self.get_fasta_repr()

    def get_gff3_string(
        self,
        chromosome=True,
        trs_type="complex_tandem_repeat",
        probability=1000,
        tool="PySatDNA",
        prefix=None,
        properties=None,
        force_header=False,
    ):
        """Return TR in gff format."""
        if chromosome and self.trf_chr and self.trf_chr != "?":
            seqid = self.trf_chr
        elif self.trf_gi and self.trf_gi != "Unknown":
            seqid = self.trf_gi
        else:
            seqid = self.trf_head
        features = []
        if not properties:
            properties = {}
        for name, attr in properties.items():
            features.append("%s=%s" % (name, getattr(self, attr)))
        features = ";".join(features)
        if prefix:
            seqid = prefix + seqid
        if self.trf_l_ind < self.trf_r_ind:
            strand = "+"
        else:
            strand = "-"
            self.trf_l_ind, self.trf_r_ind = self.trf_r_ind, self.trf_l_ind

        if force_header:
            seqid = self.trf_head
        d = (
            seqid,
            tool,
            trs_type,
            self.trf_l_ind,
            self.trf_r_ind,
            probability,
            strand,
            ".",
            features,
        )
        return "%s\n" % "\t".join(map(str, d))

    def get_bed_string(self):
        """Return TR in bed format."""
        if self.trf_l_ind < self.trf_r_ind:
            strand = "+"
        else:
            strand = "-"
            self.trf_l_ind, self.trf_r_ind = self.trf_r_ind, self.trf_l_ind

        seqid = self.trf_head
        d = (
            seqid,
            self.trf_l_ind,
            self.trf_r_ind,
        )
        return "%s\n" % "\t".join(map(str, d))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @created: 20.02.2024
# @author: Aleksey Komissarov
# @contact: ad3002@gmail.com

from .sequences import get_revcomp
from collections import Counter
from tqdm import tqdm

def best_kmer_for_start(arrays, k=5):
    k = 5
    kmers = Counter()
    for array in arrays:
        repeat = "".join(array)
        prev_kmer = None
        for i in range(len(repeat)-k+1):
            kmer = repeat[i:i+k]
            if kmer == prev_kmer:
                continue
            prev_kmer = kmer
            kmers[kmer] += 1

    best_kmer = None
    max_f1 = 0
    for kmer, _ in tqdm(kmers.most_common(100)):
        copy_number = []
        for array in arrays:
            for repeat in array:
                copy_number.append(repeat.count(kmer))

        n = len(copy_number)
        r = len([x for x in copy_number if x > 0])/n
        p = len([x for x in copy_number if x == 1])/n
        f1 = 2 * (r * p) / (r + p)
        if f1 < 0.5:
            continue
        if f1 > max_f1:
            best_kmer = (r, p, f1, kmer)
            max_f1 = f1
    return max_f1, best_kmer

def rotate_arrays(arrays_, starting_kmer=None):

    if not starting_kmer:
        arrays = []
        for array in arrays_:
            feature1 = array.count("C") < array.count("G")
            feature2 = array.count("A") < array.count("T")
            if feature1 and feature2 or (not feature1 and not feature2):
                if feature1 and feature2:
                    arrays.append(array.split())
                else:
                    arrays.append(get_revcomp(array).split())
            else:
                if feature1 and not feature2:
                    arrays.append(get_revcomp(array).split())
                else:
                    arrays.append(array.split())
        max_f1, (r, p, f1, best_kmer) = best_kmer_for_start(arrays)
    else:
        best_kmer = starting_kmer
        best_rev_kmer = get_revcomp(best_kmer)
        arrays = []
        for array in arrays_:
            forward_feature = array.count(best_rev_kmer) < array.count(best_kmer)
            if forward_feature:
                arrays.append(array.split())
            else:
                arrays.append(get_revcomp(array).split())
        
    new_arrays = []
    for array in arrays:
        new_array = []
        prev_repeat = ""
        queue = array[::]
        while queue:
            repeat = prev_repeat + queue.pop(0)
            pos = repeat.find(best_kmer, 1)
            if pos == -1:
                prev_repeat = repeat
                continue
            new_array.append(repeat[:pos])
            prev_repeat = repeat[pos:]
        if prev_repeat:
            while True:
                pos = prev_repeat.find(best_kmer, 1)
                if pos == -1:
                    new_array.append(prev_repeat)
                    break
                new_array.append(prev_repeat[:pos])
                prev_repeat = prev_repeat[pos:]
            
        assert "".join(array) == "".join(new_array)
        
        new_arrays.append(new_array)

    return new_arrays



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Tests for the frequency suffix tree module.
"""

import pytest
from ArraySplitter.core_functions.tools.fs_tree import (
    WeightedValueHeap,
    update,
    iter_fs_tree_from_sequence,
    build_fs_tree_from_sequence,
)
from ArraySplitter.decompose import (
    get_fs_tree,
    iterate_hints,
)


class TestWeightedValueHeap:
    """Test the WeightedValueHeap data structure."""
    
    def test_init(self):
        """Test heap initialization."""
        heap = WeightedValueHeap()
        assert heap.heap == []
        assert heap.peek() is None
        assert heap.pop() is None
    
    def test_insert_and_pop(self):
        """Test insertion and popping maintains heap order."""
        heap = WeightedValueHeap()
        
        # Insert values with weights
        heap.insert(5, "five")
        heap.insert(1, "one")
        heap.insert(3, "three")
        heap.insert(2, "two")
        
        # Pop should return in order of weights (min first)
        assert heap.pop() == (1, "one")
        assert heap.pop() == (2, "two")
        assert heap.pop() == (3, "three")
        assert heap.pop() == (5, "five")
        assert heap.pop() is None
    
    def test_peek(self):
        """Test peeking without removing."""
        heap = WeightedValueHeap()
        heap.insert(3, "three")
        heap.insert(1, "one")
        
        # Peek should return min without removing
        assert heap.peek() == (1, "one")
        assert heap.peek() == (1, "one")  # Still there
        assert len(heap.heap) == 2


class TestFSTreeFunctions:
    """Test frequency suffix tree construction functions."""
    
    def test_get_starting_index_simple(self):
        """Test finding starting indices for a nucleotide."""
        sequence = "ATCGATCG"
        
        # Find all 'A' positions
        indices = get_starting_index(sequence, 'A')
        assert indices == [0, 4]
        
        # Find all 'T' positions
        indices = get_starting_index(sequence, 'T')
        assert indices == [1, 5]
        
        # Find all 'G' positions
        indices = get_starting_index(sequence, 'G')
        assert indices == [3, 7]
        
        # Non-existent nucleotide
        indices = get_starting_index(sequence, 'N')
        assert indices == []
    
    def test_get_starting_index_repeated(self):
        """Test with repeated nucleotides."""
        sequence = "AAATTTCCC"
        
        indices = get_starting_index(sequence, 'A')
        assert indices == [0, 1, 2]
        
        indices = get_starting_index(sequence, 'T')
        assert indices == [3, 4, 5]
    
    @pytest.mark.parametrize("sequence,nucl,expected_count", [
        ("ATCG", "A", 1),
        ("AAAA", "A", 4),
        ("CGCG", "T", 0),
        ("", "A", 0),
    ])
    def test_get_starting_index_parametrized(self, sequence, nucl, expected_count):
        """Parametrized test for various sequences."""
        indices = get_starting_index(sequence, nucl)
        assert len(indices) == expected_count
    
    def test_get_nodes(self):
        """Test node extraction from fs_tree."""
        # Create a simple fs_tree structure
        fs_tree = {
            0: (0, ['A'], [0, 4], [1, 5], None, [1, 2]),
            1: (1, ['A', 'T'], [0], [2], 0, []),
            2: (2, ['A', 'C'], [4], [6], 0, []),
        }
        
        nodes = get_nodes(fs_tree, 0)
        assert len(nodes) == 3
        assert 0 in nodes
        assert 1 in nodes
        assert 2 in nodes
    
    def test_update_function(self):
        """Test the update function for fs_tree construction."""
        fs_tree = {0: (0, [], [0], [0], None, [])}
        queue = []
        
        # Test adding a new node
        cid = update(
            fs_tree=fs_tree,
            queue=queue,
            nucl='A',
            fs_x=[0, 1, 2, 3],  # 4 occurrences
            fs_xp=[1, 2, 3, 4],
            current_cid=0,
            cid=1,
            cutoff=3  # Will pass since len(fs_x) > cutoff
        )
        
        assert cid == 2  # Should increment
        assert 1 in fs_tree
        assert fs_tree[1][1] == ['A']
        assert fs_tree[1][2] == [0, 1, 2, 3]
        assert len(queue) == 1
        assert queue[0] == 1
    
    def test_update_below_cutoff(self):
        """Test update when frequency is below cutoff."""
        fs_tree = {0: (0, [], [0], [0], None, [])}
        queue = []
        
        cid = update(
            fs_tree=fs_tree,
            queue=queue,
            nucl='A',
            fs_x=[0, 1],  # Only 2 occurrences
            fs_xp=[1, 2],
            current_cid=0,
            cid=1,
            cutoff=3  # Won't pass since len(fs_x) <= cutoff
        )
        
        assert cid == 1  # Should not increment
        assert 1 not in fs_tree
        assert len(queue) == 0


class TestFSTreeConstruction:
    """Test the main fs_tree construction function."""
    
    def test_simple_repeat_pattern(self):
        """Test fs_tree construction on a simple repeat."""
        sequence = "ATCATCATC"
        
        # Build fs_tree starting from 'A'
        fs_tree = get_fs_tree(sequence, 'A', cutoff=1, depth=3)
        
        # Should find the ATC pattern
        assert len(fs_tree) > 1
        
        # Check that starting positions are correct
        root_node = fs_tree[0]
        assert root_node[2] == [0, 3, 6]  # Positions of 'A'
    
    def test_no_pattern(self):
        """Test with sequence without repeating patterns."""
        sequence = "ACGTACGT"
        
        fs_tree = get_fs_tree(sequence, 'A', cutoff=3, depth=3)
        
        # With high cutoff, should only have root
        assert len(fs_tree) == 1


@pytest.fixture
def sample_tandem_repeat():
    """Provide a sample tandem repeat sequence."""
    return "CAGCAGCAGCAGCAG"  # CAG repeat


@pytest.fixture
def complex_satellite():
    """Provide a more complex satellite array."""
    monomer = "ATCGATCGATCG"
    return monomer * 10  # 10 copies


def test_iterate_hints(sample_tandem_repeat):
    """Test hint generation from fs_tree."""
    fs_tree = get_fs_tree(sample_tandem_repeat, 'C', cutoff=2, depth=5)
    hints = list(iterate_hints(fs_tree, 0))
    
    # Should find CAG pattern
    assert len(hints) > 0
    
    # Check if CAG is among the hints
    hint_sequences = [hint[1] for hint in hints]
    assert any('CAG' in seq for seq in hint_sequences)


"""
Anchor Graph for monomer decomposition.

Key idea: Build a graph where:
- Nodes = conserved parts (anchors) from FS-tree hints
- Edges = transitions between anchors with support (copy number)

This solves two problems:
1. Multiple conserved parts per monomer (overcutting)
2. Mutations in anchor (missing monomers) - handled by using shorter anchors
   where longer ones are absent

The graph structure reveals the monomer architecture.
"""

from collections import defaultdict, Counter
from typing import List, Tuple, Dict, Set, Optional
from dataclasses import dataclass, field
from statistics import median


@dataclass
class AnchorHit:
    """Single occurrence of an anchor in sequence."""
    anchor: str
    position: int
    length: int

    def __lt__(self, other):
        return self.position < other.position


def get_top_candidates(candidates: List[Dict], top_k: int = 10) -> List[Dict]:
    """
    Get top-k candidates from compute_cuts output.

    Candidates are already scored by compute_cuts logic.
    """
    # Sort by adjusted_score (or base_score if adjusted not available)
    sorted_candidates = sorted(
        candidates,
        key=lambda x: x.get('adjusted_score', x.get('base_score', 0)),
        reverse=True
    )
    return sorted_candidates[:top_k]


# =============================================================================
# Distance Clustering Functions
# =============================================================================

def simple_kmeans(data: List[float], k: int, max_iter: int = 100) -> Tuple[List[List[float]], List[float]]:
    """
    Simple k-means clustering without external dependencies.

    Args:
        data: List of values to cluster
        k: Number of clusters
        max_iter: Maximum iterations

    Returns:
        (clusters, centers) - list of clusters and their centers
    """
    if not data or k <= 0:
        return [[]], [0]

    if k == 1:
        return [data], [sum(data) / len(data)]

    # Initialize centers evenly across the data range
    data_sorted = sorted(data)
    n = len(data_sorted)
    centers = [data_sorted[min(i * n // k, n - 1)] for i in range(k)]

    for _ in range(max_iter):
        # Assign points to nearest center
        clusters = [[] for _ in range(k)]
        for x in data:
            nearest = min(range(k), key=lambda i: abs(x - centers[i]))
            clusters[nearest].append(x)

        # Update centers
        new_centers = []
        for i, c in enumerate(clusters):
            if c:
                new_centers.append(sum(c) / len(c))
            else:
                # Empty cluster - keep old center
                new_centers.append(centers[i])

        if new_centers == centers:
            break
        centers = new_centers

    return clusters, centers


def find_optimal_clusters(data: List[float], max_k: int = 4) -> Tuple[int, List[List[float]], List[float]]:
    """
    Find optimal number of clusters (1 to max_k).

    Criteria:
    - All clusters must be significant (>15% of data)
    - Cluster centers must be well separated (gap > 20% of max center)

    Returns:
        (n_clusters, clusters, centers)
    """
    if len(data) < 4:
        return 1, [data], [median(data)]

    best_k = 1
    best_score = 0
    best_clusters = [data]
    best_centers = [median(data)]

    for k in range(2, min(max_k + 1, len(data) // 2 + 1)):
        clusters, centers = simple_kmeans(data, k)

        # Filter out empty clusters
        non_empty = [(c, ctr) for c, ctr in zip(clusters, centers) if c]
        if len(non_empty) < k:
            continue

        clusters = [x[0] for x in non_empty]
        centers = [x[1] for x in non_empty]

        # Check that all clusters are significant (>15% of data)
        min_size = min(len(c) for c in clusters)
        if min_size < len(data) * 0.15:
            continue

        # Check that centers are well separated
        centers_sorted = sorted(centers)
        if len(centers_sorted) < 2:
            continue

        min_gap = min(centers_sorted[i + 1] - centers_sorted[i]
                      for i in range(len(centers_sorted) - 1))
        max_center = max(centers_sorted)

        if min_gap < 0.2 * max_center:
            continue

        # Score: separation quality
        score = min_gap / max_center
        if score > best_score:
            best_score = score
            best_k = k
            best_clusters = clusters
            best_centers = centers

    return best_k, best_clusters, best_centers


def analyze_distance_distribution(distances: List[float], verbose: bool = False) -> Tuple[int, List[float], float]:
    """
    Analyze distance distribution to detect multi-region patterns.

    If anchor appears N times per monomer, distances will have N clusters.
    True period = sum of cluster centers.

    Args:
        distances: List of distances between consecutive anchor occurrences
        verbose: Print debug info

    Returns:
        (n_clusters, cluster_centers, true_period)
    """
    if len(distances) < 4:
        med = median(distances) if distances else 0
        return 1, [med], med

    n_clusters, clusters, centers = find_optimal_clusters(distances, max_k=4)

    # Sort centers by value for consistent ordering
    centers_sorted = sorted(centers)
    true_period = sum(centers_sorted)

    if verbose and n_clusters > 1:
        print(f"  Distance clustering: {n_clusters} clusters")
        for i, (c, ctr) in enumerate(sorted(zip(clusters, centers), key=lambda x: x[1])):
            print(f"    Cluster {i}: center={ctr:.0f}, count={len(c)}")
        print(f"  True period: {true_period:.0f}")

    return n_clusters, centers_sorted, true_period


def classify_positions_by_distance(
    positions: List[int],
    centers: List[float],
    verbose: bool = False
) -> List[int]:
    """
    Classify each position by the cluster of its distance to the next position.

    Args:
        positions: Sorted list of anchor positions
        centers: Cluster centers from analyze_distance_distribution

    Returns:
        List of cluster labels (0, 1, 2, ...) for each position.
        Last position gets label -1 (no next position).
    """
    if len(positions) < 2:
        return [-1] * len(positions)

    labels = []
    for i in range(len(positions) - 1):
        dist = positions[i + 1] - positions[i]
        # Find nearest cluster center
        nearest = min(range(len(centers)), key=lambda j: abs(dist - centers[j]))
        labels.append(nearest)

    # Last position has no "next", mark with -1
    labels.append(-1)

    if verbose:
        label_counts = Counter(l for l in labels if l >= 0)
        print(f"  Position labels: {dict(label_counts)}")

    return labels


def select_cut_positions_by_cluster(
    positions: List[int],
    labels: List[int],
    verbose: bool = False
) -> List[int]:
    """
    Select cut positions based on cluster analysis.

    Strategy: Cut at positions with the same label as the first position.
    This preserves the monomer structure starting from the left flank.

    Args:
        positions: Sorted list of anchor positions
        labels: Cluster labels for each position

    Returns:
        List of positions to cut at
    """
    if not positions or not labels:
        return positions

    # Use the label of the first position as the "start" label
    # This defines which region (A, B, C...) is the monomer start
    start_label = labels[0]

    if start_label < 0:
        # Only one position, return it
        return positions

    # Select all positions with the start label
    cut_positions = [pos for pos, label in zip(positions, labels) if label == start_label]

    # Also include the last position if it wasn't included
    # (to capture the final partial monomer)
    if positions[-1] not in cut_positions:
        cut_positions.append(positions[-1])

    if verbose:
        print(f"  Start label: {start_label}")
        print(f"  Cut positions: {len(cut_positions)} of {len(positions)} total")

    return sorted(cut_positions)


def find_anchor_hits_hierarchical(
    sequence: str,
    anchors: List[str],
    verbose: bool = False
) -> List[AnchorHit]:
    """
    Find all anchor occurrences, using longer anchors first.

    Logic:
    1. Sort anchors by length (longest first)
    2. Find all positions of longest anchor
    3. For shorter anchors - only find in regions NOT covered by longer ones
    4. This handles mutations: if long anchor has mutation, short variant will be found

    Args:
        sequence: The array sequence
        anchors: List of anchor sequences (will be sorted by length)
        verbose: Print debug info

    Returns:
        List of AnchorHit sorted by position
    """
    if not anchors:
        return []

    # Sort by length descending
    sorted_anchors = sorted(anchors, key=len, reverse=True)

    # Track covered positions
    covered = set()
    all_hits = []

    for anchor in sorted_anchors:
        anchor_len = len(anchor)

        # Find all occurrences
        start = 0
        hits_for_anchor = 0

        while True:
            pos = sequence.find(anchor, start)
            if pos == -1:
                break

            # Check if this position is already covered by a longer anchor
            # A position is "covered" if ANY part of this anchor overlaps with covered region
            is_covered = any(p in covered for p in range(pos, pos + anchor_len))

            if not is_covered:
                all_hits.append(AnchorHit(anchor=anchor, position=pos, length=anchor_len))
                # Mark positions as covered
                for p in range(pos, pos + anchor_len):
                    covered.add(p)
                hits_for_anchor += 1

            start = pos + 1

        if verbose and hits_for_anchor > 0:
            print(f"  '{anchor}' (len={anchor_len}): {hits_for_anchor} hits")

    # Sort by position
    all_hits.sort()

    return all_hits


def build_transition_graph(hits: List[AnchorHit]) -> Dict:
    """
    Build graph of transitions between consecutive anchor hits.

    Returns dict with:
    - edges: {(from_anchor, to_anchor): count}
    - distances: {(from_anchor, to_anchor): [list of distances]}
    - sequences: {(from_anchor, to_anchor): [list of sequences between]}
    """
    edges = Counter()
    distances = defaultdict(list)

    for i in range(len(hits) - 1):
        curr = hits[i]
        next_hit = hits[i + 1]

        edge = (curr.anchor, next_hit.anchor)
        distance = next_hit.position - curr.position

        edges[edge] += 1
        distances[edge].append(distance)

    return {
        "edges": dict(edges),
        "distances": dict(distances),
    }


def find_monomer_cycle(graph: Dict, hits: List[AnchorHit], sequence: str, verbose: bool = False) -> Tuple[List[str], float, float, int, List[float]]:
    """
    Find the best anchor for decomposition using distance clustering.

    Strategy:
    1. For each anchor with self-loop, analyze distance distribution
    2. Cluster distances to detect multi-region patterns (A-B, A-B-C, etc.)
    3. True period = sum of cluster centers
    4. Select anchor with lowest CV after cluster-based correction

    Returns:
        Tuple of (cycle, mean_period, cv, n_clusters, cluster_centers)
    """
    edges = graph["edges"]
    distances = graph.get("distances", {})

    if not edges:
        return [], 0, float('inf'), 1, []

    # Collect candidate anchors with cluster analysis
    # Format: (cycle, cv, mean_period, count, n_clusters, centers)
    candidates = []

    # Get unique anchors from graph with self-loops
    anchors_to_try = set()
    for (from_a, to_a), count in edges.items():
        if count >= 3:  # Need enough occurrences
            anchors_to_try.add(from_a)
            if from_a != to_a:
                anchors_to_try.add(to_a)

    for anchor in anchors_to_try:
        # Find ALL positions for this anchor (not just self-loop)
        positions = [hit.position for hit in hits if hit.anchor == anchor]
        if len(positions) < 3:
            continue
        positions.sort()

        # Calculate distances between ALL consecutive positions of this anchor
        # This captures the full pattern even if other anchors appear in between
        dist_list = []
        for i in range(len(positions) - 1):
            dist_list.append(positions[i + 1] - positions[i])

        if len(dist_list) < 3:
            continue

        # Analyze distance distribution - detect multi-region patterns
        n_clusters, centers, true_period = analyze_distance_distribution(dist_list, verbose=False)

        # Classify positions and get cut positions
        labels = classify_positions_by_distance(positions, centers)
        cut_positions = select_cut_positions_by_cluster(positions, labels)

        if len(cut_positions) < 2:
            continue

        # Calculate CV of resulting monomer lengths
        lengths = []
        for i in range(len(cut_positions) - 1):
            lengths.append(cut_positions[i + 1] - cut_positions[i])

        if len(lengths) < 2:
            continue

        mean_len = sum(lengths) / len(lengths)
        variance = sum((x - mean_len) ** 2 for x in lengths) / len(lengths)
        cv = (variance ** 0.5) / mean_len if mean_len > 0 else float('inf')

        candidates.append(([anchor], cv, true_period, len(cut_positions), n_clusters, centers))

        if verbose:
            cluster_info = f" ({n_clusters} clusters)" if n_clusters > 1 else ""
            print(f"  Candidate: {anchor[:20]}... period={true_period:.0f}, CV={cv:.3f}, cuts={len(cut_positions)}{cluster_info}")

    # Select best candidate by CV
    if not candidates:
        # Fallback: most frequent edge
        if edges:
            start_edge = max(edges.items(), key=lambda x: x[1])[0]
            if verbose:
                print(f"  Fallback to most frequent: {start_edge[0][:20]}...")
            return [start_edge[0]], 0, float('inf'), 1, []
        return [], 0, float('inf'), 1, []

    # Sort by CV (ascending), then by n_clusters (ascending) - prefer simpler patterns
    candidates.sort(key=lambda x: (x[1], x[4]))

    best_cycle, best_cv, best_period, best_count, best_n_clusters, best_centers = candidates[0]

    if verbose:
        cluster_info = f" ({best_n_clusters} clusters)" if best_n_clusters > 1 else ""
        print(f"  Selected: {best_cycle[0][:20]}... period={best_period:.0f}, CV={best_cv:.3f}{cluster_info}")
        if len(candidates) > 1:
            print(f"  (rejected {len(candidates)-1} other candidates)")

    return best_cycle, best_period, best_cv, best_n_clusters, best_centers


def _calculate_cycle_distances(cycle: List[str], distances: Dict) -> List[float]:
    """Calculate full cycle distances from edge distances."""
    if len(cycle) == 1:
        edge = (cycle[0], cycle[0])
        return distances.get(edge, [])

    # For multi-anchor cycle, sum distances along the cycle
    # This is approximate - we'd need to track actual cycle instances
    cycle_dists = []
    edge = (cycle[0], cycle[1])
    if edge in distances:
        # Use first edge distances as proxy
        # TODO: properly track full cycle instances
        cycle_dists = distances[edge]

    return cycle_dists


def estimate_monomer_length(graph: Dict, cycle: List[str]) -> Tuple[float, float]:
    """
    Estimate monomer length from cycle and transition distances.

    Returns (mean_length, variance)
    """
    if not cycle:
        return 0, 0

    distances = graph["distances"]

    # Sum distances along the cycle
    cycle_lengths = []

    # We need to find instances where the full cycle occurs
    # For now, estimate from individual edge distances
    total_distances = []

    for i in range(len(cycle)):
        from_anchor = cycle[i]
        to_anchor = cycle[(i + 1) % len(cycle)]
        edge = (from_anchor, to_anchor)

        if edge in distances:
            total_distances.extend(distances[edge])

    if not total_distances:
        return 0, 0

    # Rough estimate: sum of mean distances along cycle
    edge_means = []
    for i in range(len(cycle)):
        from_anchor = cycle[i]
        to_anchor = cycle[(i + 1) % len(cycle)]
        edge = (from_anchor, to_anchor)

        if edge in distances:
            edge_means.append(sum(distances[edge]) / len(distances[edge]))

    if edge_means:
        estimated_length = sum(edge_means)
        variance = sum((d - estimated_length/len(cycle))**2 for d in total_distances) / len(total_distances)
        return estimated_length, variance

    return 0, 0


def decompose_by_cycle(
    sequence: str,
    hits: List[AnchorHit],
    cycle: List[str],
    cluster_centers: List[float] = None,
    verbose: bool = False
) -> List[str]:
    """
    Decompose sequence into monomers based on the anchor cycle.

    Strategy:
    1. Find all positions of the anchor
    2. If cluster_centers provided (multi-region pattern), use cluster analysis
       to select only the "start" positions
    3. Cut at selected positions

    Args:
        sequence: The array sequence
        hits: Sorted list of AnchorHit
        cycle: List of anchors representing one monomer
        cluster_centers: If multi-region pattern detected, the cluster centers
                        for classifying positions. None = cut at every position.

    Returns:
        List of monomer sequences
    """
    if not cycle or not hits:
        return [sequence] if sequence else []

    start_anchor = cycle[0]

    # Find all positions where anchor appears
    all_positions = []
    for hit in hits:
        if hit.anchor == start_anchor:
            all_positions.append(hit.position)

    if not all_positions:
        return [sequence]

    all_positions.sort()

    # Determine cut positions based on cluster analysis
    if cluster_centers and len(cluster_centers) > 1:
        # Multi-region pattern: use cluster-based selection
        labels = classify_positions_by_distance(all_positions, cluster_centers)
        cut_positions = select_cut_positions_by_cluster(all_positions, labels, verbose)

        if verbose:
            print(f"  Multi-region pattern: {len(cluster_centers)} clusters")
            print(f"  All positions: {len(all_positions)}, cut positions: {len(cut_positions)}")
    else:
        # Single region: cut at every position
        cut_positions = all_positions

    if verbose:
        print(f"  Cut positions: {cut_positions[:10]}{'...' if len(cut_positions) > 10 else ''}")

    if not cut_positions:
        return [sequence]

    # Cut sequence
    monomers = []

    # Left flank
    if cut_positions[0] > 0:
        monomers.append(sequence[:cut_positions[0]])

    # Monomers
    for i in range(len(cut_positions) - 1):
        monomers.append(sequence[cut_positions[i]:cut_positions[i + 1]])

    # Last part (may be partial)
    if cut_positions[-1] < len(sequence):
        monomers.append(sequence[cut_positions[-1]:])

    return monomers


class AnchorGraphDecomposer:
    """
    Main class for anchor graph-based decomposition.

    Usage:
        decomposer = AnchorGraphDecomposer()
        decomposer.build_from_candidates(sequence, candidates)
        monomers = decomposer.decompose()
    """

    def __init__(self):
        self.sequence: Optional[str] = None
        self.anchors: List[str] = []
        self.hits: List[AnchorHit] = []
        self.graph: Dict = {}
        self.cycle: List[str] = []
        self.estimated_period: float = 0
        self.estimated_cv: float = float('inf')
        self.n_clusters: int = 1  # Number of distance clusters (1 = normal, >1 = multi-region)
        self.cluster_centers: List[float] = []  # Cluster centers for multi-region patterns

    def build_from_candidates(
        self,
        sequence: str,
        candidates: List[Dict],
        top_k: int = 10,
        verbose: bool = False
    ) -> None:
        """
        Build anchor graph from compute_cuts candidates.

        Args:
            sequence: The array sequence
            candidates: List of candidate dicts from compute_cuts
            top_k: Number of top candidates to use
            verbose: Print debug info
        """
        self.sequence = sequence

        # Get top candidates, filtering by max anchor length
        # Anchors should be short (<=11bp) - they are graph nodes, not sequences
        MAX_ANCHOR_LENGTH = 11
        top = get_top_candidates(candidates, top_k)
        self.anchors = [c['cut'] for c in top if len(c['cut']) <= MAX_ANCHOR_LENGTH]

        if verbose:
            print(f"Top {len(self.anchors)} anchors (max length {MAX_ANCHOR_LENGTH}bp):")
            for c in top:
                if len(c['cut']) <= MAX_ANCHOR_LENGTH:
                    print(f"  '{c['cut']}' score={c.get('adjusted_score', c.get('base_score', 0)):.3f}")

        # Find hits hierarchically (longer first)
        if verbose:
            print(f"\nFinding anchor hits (longer anchors first):")
        self.hits = find_anchor_hits_hierarchical(sequence, self.anchors, verbose)

        if verbose:
            print(f"\nTotal hits: {len(self.hits)}")

        # Build transition graph
        self.graph = build_transition_graph(self.hits)

        if verbose:
            print(f"\nTransition graph:")
            for edge, count in sorted(self.graph["edges"].items(), key=lambda x: x[1], reverse=True)[:10]:
                dists = self.graph["distances"][edge]
                mean_dist = sum(dists) / len(dists)
                print(f"  {edge[0][:15]}... -> {edge[1][:15]}...: count={count}, mean_dist={mean_dist:.0f}")

        # Find monomer cycle with cluster analysis
        if verbose:
            print(f"\nFinding monomer cycle:")
        self.cycle, self.estimated_period, self.estimated_cv, self.n_clusters, self.cluster_centers = find_monomer_cycle(
            self.graph, self.hits, sequence, verbose
        )

        if verbose and self.cycle:
            cluster_info = f", {self.n_clusters} clusters" if self.n_clusters > 1 else ""
            print(f"\nEstimated monomer length: {self.estimated_period:.0f} bp (CV={self.estimated_cv:.3f}{cluster_info})")

    def decompose(self, verbose: bool = False) -> List[str]:
        """
        Decompose sequence into monomers.

        Returns:
            List of monomer sequences
        """
        if not self.cycle:
            if verbose:
                print("No cycle found, returning whole sequence")
            return [self.sequence] if self.sequence else []

        if verbose:
            cluster_info = f" ({self.n_clusters} clusters)" if self.n_clusters > 1 else ""
            print(f"\nDecomposing by cycle: {' -> '.join(a[:10]+'...' for a in self.cycle)}{cluster_info}")

        monomers = decompose_by_cycle(self.sequence, self.hits, self.cycle, self.cluster_centers, verbose)

        # Verify reconstruction
        reconstructed = "".join(monomers)
        if reconstructed != self.sequence:
            if verbose:
                print(f"WARNING: Reconstruction mismatch! {len(self.sequence)} vs {len(reconstructed)}")
        elif verbose:
            print(f"Reconstruction: PERFECT")

        return monomers

    def get_stats(self) -> Dict:
        """Get decomposer statistics."""
        return {
            "num_anchors": len(self.anchors),
            "anchors": self.anchors,
            "num_hits": len(self.hits),
            "cycle": self.cycle,
            "estimated_monomer_length": self.estimated_period,
            "estimated_cv": self.estimated_cv,
            "n_clusters": self.n_clusters,
            "cluster_centers": self.cluster_centers,
            "edge_counts": self.graph.get("edges", {}),
        }



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Tests for sequence utility functions.
"""

import pytest
from ArraySplitter.core_functions.tools.sequences import (
    get_revcomp,
    clear_sequence,
)


class TestReverseComplement:
    """Test reverse complement function."""
    
    def test_get_revcomp_simple(self):
        """Test basic reverse complement."""
        assert get_revcomp("ATCG") == "CGAT"
        assert get_revcomp("AAAA") == "TTTT"
        assert get_revcomp("TTTT") == "AAAA"
        assert get_revcomp("CCCC") == "GGGG"
        assert get_revcomp("GGGG") == "CCCC"
    
    def test_get_revcomp_mixed(self):
        """Test with mixed sequences."""
        assert get_revcomp("ATCGATCG") == "CGATCGAT"
        assert get_revcomp("GCTAGCTA") == "TAGCTAGC"
    
    def test_get_revcomp_palindrome(self):
        """Test palindromic sequences."""
        # These should be self-complementary when reversed
        assert get_revcomp("ATCGAT") == "ATCGAT"
        assert get_revcomp("GCATGC") == "GCATGC"
    
    def test_get_revcomp_empty(self):
        """Test empty sequence."""
        assert get_revcomp("") == ""
    
    def test_get_revcomp_lowercase(self):
        """Test handling of lowercase."""
        assert get_revcomp("atcg") == "cgat"
        assert get_revcomp("AtCg") == "CgAt"
    
    @pytest.mark.parametrize("seq,expected", [
        ("A", "T"),
        ("T", "A"),
        ("C", "G"),
        ("G", "C"),
        ("N", "N"),  # Ambiguous bases
        ("ATCGN", "NCGAT"),
    ])
    def test_get_revcomp_parametrized(self, seq, expected):
        """Parametrized tests for various sequences."""
        assert get_revcomp(seq) == expected


class TestClearSequence:
    """Test sequence cleaning function."""
    
    def test_clear_sequence_basic(self):
        """Test basic sequence cleaning."""
        assert clear_sequence("ATCG") == "ATCG"
        assert clear_sequence("atcg") == "ATCG"
        assert clear_sequence("AtCg") == "ATCG"
    
    def test_clear_sequence_with_spaces(self):
        """Test removal of spaces."""
        assert clear_sequence("A T C G") == "ATCG"
        assert clear_sequence("  ATCG  ") == "ATCG"
        assert clear_sequence("A\tT\nC\rG") == "ATCG"
    
    def test_clear_sequence_with_numbers(self):
        """Test removal of numbers."""
        assert clear_sequence("1ATCG") == "ATCG"
        assert clear_sequence("ATCG123") == "ATCG"
        assert clear_sequence("A1T2C3G4") == "ATCG"
    
    def test_clear_sequence_fasta_header(self):
        """Test cleaning FASTA-like input."""
        assert clear_sequence(">seq1\nATCG") == "ATCG"
        assert clear_sequence("ATCG\n>seq2\nGCTA") == "ATCGGCTA"
    
    def test_clear_sequence_empty(self):
        """Test empty and whitespace-only sequences."""
        assert clear_sequence("") == ""
        assert clear_sequence("   ") == ""
        assert clear_sequence("\n\n\n") == ""
    
    def test_clear_sequence_special_chars(self):
        """Test removal of special characters."""
        assert clear_sequence("A-T-C-G") == "ATCG"
        assert clear_sequence("A.T.C.G") == "ATCG"
        assert clear_sequence("A*T*C*G") == "ATCG"
        assert clear_sequence("(ATCG)") == "ATCG"
    
    def test_clear_sequence_ambiguous_bases(self):
        """Test handling of ambiguous bases."""
        # Assuming the function keeps standard IUPAC codes
        result = clear_sequence("ATCGRYSWKMBDHVN")
        assert "ATCG" in result
        # Check if ambiguous bases are handled (kept or removed)
        assert len(result) >= 4
    
    @pytest.mark.parametrize("input_seq,expected", [
        ("atcg", "ATCG"),
        ("ATCG", "ATCG"),
        ("a t c g", "ATCG"),
        ("1a2t3c4g5", "ATCG"),
        (">header\natcg", "ATCG"),
        ("", ""),
    ])
    def test_clear_sequence_parametrized(self, input_seq, expected):
        """Parametrized tests for sequence cleaning."""
        assert clear_sequence(input_seq) == expected


class TestSequenceUtilities:
    """Test combined sequence operations."""
    
    def test_clean_and_revcomp(self):
        """Test cleaning followed by reverse complement."""
        dirty_seq = "  atcg 123  "
        cleaned = clear_sequence(dirty_seq)
        assert cleaned == "ATCG"
        assert get_revcomp(cleaned) == "CGAT"
    
    def test_revcomp_idempotent(self):
        """Test that double reverse complement returns original."""
        original = "ATCGATCGATCG"
        revcomp1 = get_revcomp(original)
        revcomp2 = get_revcomp(revcomp1)
        assert revcomp2 == original
    
    def test_clear_idempotent(self):
        """Test that cleaning twice gives same result."""
        dirty = "  atcg 123  "
        clean1 = clear_sequence(dirty)
        clean2 = clear_sequence(clean1)
        assert clean1 == clean2


@pytest.fixture
def messy_fasta_content():
    """Provide messy FASTA-like content."""
    return """
    >sequence_1 description
    ATCGATCG atcg
    123 GCTAGCTA
    
    >sequence_2
    aaaa tttt
    cccc gggg
    """


def test_clear_multiline_fasta(messy_fasta_content):
    """Test cleaning multi-line FASTA content."""
    # This might need adjustment based on actual function behavior
    cleaned = clear_sequence(messy_fasta_content)
    # Should contain all sequences without headers or formatting
    assert "ATCGATCG" in cleaned
    assert "GCTAGCTA" in cleaned
    assert "AAAA" in cleaned or "aaaa" in cleaned.lower()
    assert ">" not in cleaned
    assert " " not in cleaned
    assert "\n" not in cleaned


import re

from setuptools import find_packages, setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

version = None
for line in open("./src/ArraySplitter/__init__.py"):
    m = re.search("__version__\s*=\s*(.*)", line)
    if m:
        version = m.group(1).strip()[1:-1]  # quotes
        break
assert version

setup(
    name="ArraySplitter",
    version=version,
    packages=find_packages('src'),
    package_dir={'': 'src'},
    package_data={"": ["README.md"]},
    python_requires=">=3.6",
    include_package_data=True,
    scripts=[],
    license="BSD",
    url="https://github.com/aglabx/ArraySplitter",
    author="Aleksey Komissarov",
    author_email="ad3002@gmail.com",
    description="De Novo Decomposition of Satellite DNA Arrays into Monomers within Telomere-to-Telomere Assemblies",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        "PyExp",
        "editdistance",
        "tqdm",
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    entry_points={
        'console_scripts': [
            'arraysplitter = ArraySplitter.main:run_it',
        ],
    },
)
